// ----------------------------------------------------------------------------
// * Raima Database Manager
// *
// * Copyright (c) 2012 Raima Inc.,  All rights reserved.
// *
// * Use of this software, whether in source code format, or in executable,
// * binary object code form, is governed by the Raima LICENSE which
// * is fully described in the LICENSE.TXT file, included within this
// * distribution of files.
// * ----------------------------------------------------------------------------

#include <iostream>
#include <AclAPI.h>
#include <process.h>

// rdm.h is required in all files using the RDM Database
#include "rdm.h"

// obj15_db.h was generated from the DDL file obj15_db.ddl.
#include "obj15_db.h"

// obj15_db_api.h was generated from the DDL file obj15_db.ddl. with 
// the -cpp option
#include "obj15_db_api.h"

static uint8_t quit = 0;

// Name space for the RDM C++ interface
using namespace RDM_CPP;

// EXAMPLE - obj15 for C++
// This example shows multi-user access to a database using separate threads
// and a remote tfserver.  The connections will use TCP/IP to communicate
// with the tfserver to do their work.  After starting the tfserver, 
// two writer threads are created.  The first begins inserting values into
// the table.  The second waits until the table contains more than 100 rows
// and then begins deleting the oldest.  A reader thread is also created
// that prints out the total number of records, the max value, and the min
// value.  Each thread uses regular locking.

// The DDL (Database Definition Language) is in obj15_db.ddl. The database
// schema is embedded into the application and the database will be created
// the first time the application is run.  Subsequent runs will use the now
// existing database.

// The following RDM C++ methods are illustrated:
//      Db->Open
//      Db->Initialize
//      Db->BeginUpdate
//      Db->End
//      Db->Abort
//      Db->BeginRead
//      Db->New_recordWithFieldValues
//      Db->Get_recordsByKey
//      Cursor->AfterLast
//      Cursor->GetFieldValues
//      Cursor++
// They are fully documented in http://docs.raima.com/rdm/.

// Each of these methods can throw and an exception of type rdm_exception

// The actual database is stored in a directory named 'obj15_db' in this
// project directory.

class obj15
{
public:
    Db_obj15_db    db;

    void InsertData () throw (rdm_exception&);
    void DeleteData() throw (rdm_exception&);
    void DisplayData() throw (rdm_exception&);
    void Initialize (uint16_t init_db=0) throw (rdm_exception&);
    void Cleanup () throw (rdm_exception&);
};

void obj15::Initialize(
    uint16_t init_db)
{
    // Open the database and initialize it
    try
    {
        db = db.Open("obj15_db");
        if(init_db)
        {
            db.Initialize();
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") opening/initializing obj15_db database" << endl;
        throw (e);
        return;
    }
    return;
}

void obj15::InsertData()
{
    int32_t      ii;
    struct rec01 rec01_rec;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginUpdate(rtypes);

            rec01_rec.field01 = ii;
            db.New_rec01_recordWithFieldValues(&rec01_rec);

            db.End();
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") adding records" << endl;
        throw (e);
        return;
    }
    return;
}

void obj15::DeleteData()
{
    int32_t    ii;
    uint64_t   rec_count;
    recordType rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};
        
    try
    {
        // Loop until there is an error or we are told to quit
        for (ii=0; !quit; ii++)
        {
            db.BeginRead(rtypes);
            rec_count = db.Get_rec01_records().GetCount();
            db.End();

            if(rec_count > 100)
            {
                db.BeginUpdate(rtypes);
                db.Get_rec01_recordsBy_field01().First().Delete();
                db.End();
            }
            else
            {
                /* Wait for a bit to allow the inserter thread to get more records
                 * in the database
                 */
                Sleep(500);
            }
        }
    }
    catch (rdm_exception& e)
    {
        db.Abort();
        cerr << "Error " << e.error() << " (" << e.what() << ") deleting records" << endl;
        throw (e);
        return;
    }
    return;
}
    
void obj15::DisplayData()
{
    Cursor_rec01 rec01_cursor;
    uint64_t     rec_count = 0;
    int32_t      field01_min;
    int32_t      field01_max;
    char         msg[256];
    recordType   rtypes[]= {Cursor_rec01::Type (), Cursor::NoMoreTypes()};

    cout << endl;
    cout << "Total Records   Min Val  Max Val" << endl;
    cout << "-------------   -------  -------" << endl;

    try
    {
        // Loop until there is an error or we are told to quit
        while(!quit)
        {
            // Wait one second
            Sleep(1000);

            db.BeginRead(rtypes);

            rec01_cursor = db.Get_rec01_recordsBy_field01();

            rec_count = rec01_cursor.GetCount();

            rec01_cursor.First();
            rec01_cursor.Get_field01(&field01_min);

            rec01_cursor.Last();
            rec01_cursor.Get_field01(&field01_max);

            db.End();

            sprintf(msg, "       %6I64d    %6d   %6d", rec_count, field01_min, field01_max);
            cout << msg << endl;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") reading records" << endl;
        if(e.error() != S_UNAVAIL)
        {
            db.End();
        }
        return;
    }   
}

void obj15::Cleanup()
{
    return;
}

/* Start an external TFS Process */
int32_t start_tfs_process ()
{
    STARTUPINFO         si;
    PROCESS_INFORMATION pi;
 
    memset( &si, 0, sizeof(si) );
    si.cb = sizeof(si);
    memset( &pi, 0, sizeof(pi) );

    if( !CreateProcess( NULL,
        "Debug\\tfserver.exe -q -start",
        NULL,
        NULL,
        FALSE,
        0,
        NULL,
        "Debug",
        &si,
        &pi )
    ) 
    {
        printf( "CreateProcess failed (%d).\n", GetLastError() );
        return 1;
    }

    /* Wait until child process exits. */
    WaitForSingleObject( pi.hProcess, INFINITE );

    /* Close process and thread handles. */
    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );

    return 0;
}

/* Stop an external TFS Process */
int32_t stop_tfs_process ()
{
    STARTUPINFO         si;
    PROCESS_INFORMATION pi;
 
    memset( &si, 0, sizeof(si) );
    si.cb = sizeof(si);
    memset( &pi, 0, sizeof(pi) );

    if( !CreateProcess( NULL,
        "Debug\\tfserver.exe -q -stop",
        NULL,
        NULL,
        FALSE,
        0,
        NULL,
        NULL,
        &si,
        &pi )
    ) 
    {
        printf( "CreateProcess failed (%d).\n", GetLastError() );
        return 1;
    }

    /* Wait until child process exits. */
    WaitForSingleObject( pi.hProcess, INFINITE );

    /* Close process and thread handles. */
    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );

    return 0;
}

/* Deleter thread */
static uint32_t __stdcall DeleterThreadFcn(
    void *arg)
{
    obj15 obj;

    try
    {
        obj.Initialize();
        obj.DeleteData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in DeleterThread" << endl;
        _endthreadex( e.error() );
        return e.error();
    }
    _endthreadex( 0 );
    return 0;
}

/* Inserter thread */
static uint32_t __stdcall InserterThreadFcn(
    void *arg)
{
    obj15 obj;

    try
    {
        obj.Initialize();
        obj.InsertData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in InserterThread" << endl;
        _endthreadex( e.error() );
        return e.error();
    }
    _endthreadex( 0 );
    return 0;
}

/* Reader thread */
static uint32_t __stdcall ReaderThreadFcn(
    void *arg)
{
    obj15 obj;

    try
    {
        obj.Initialize();
        obj.DisplayData();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") in ReaderThread" << endl;
        _endthreadex( e.error() );
        return e.error();
    }
    _endthreadex( 0 );
    return 0;
}

int32_t main(void)
{
    TFS_HANDLE hTFS;
    uint32_t   inserter_tid;
    uint32_t   deleter_tid;
    uint32_t   reader_tid;
    HANDLE     hInserterThread;
    HANDLE     hDeleterThread;
    HANDLE     hReaderThread;
    obj15      obj;

    /* Start an external TFS process */
    if ( start_tfs_process() != 0)
    {
        fprintf(stderr, "There was an error starting the external TFS process\n");
        return 1;
    }

    /* Initialize the application to use the external TFS (RPC) that we 
     * started */
    d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

    try
    {
        obj.Initialize(1);
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ")" << endl;
        return EXIT_FAILURE;
    }

    cout << endl << "Starting Threads..." << endl;

    hInserterThread = (HANDLE) _beginthreadex(NULL, 0, InserterThreadFcn, NULL, 0, &inserter_tid);
    hDeleterThread  = (HANDLE) _beginthreadex(NULL, 0, DeleterThreadFcn, NULL, 0, &deleter_tid);
    hReaderThread   = (HANDLE) _beginthreadex(NULL, 0, ReaderThreadFcn, NULL, 0, &reader_tid);

    cout << "Press <Enter> when you are ready to terminate them . . ." << endl;
    getchar();

    /* Tell the threads to quit */    
    quit = 1;

    /* Wait for the reader thread */
    WaitForSingleObject(hReaderThread, INFINITE);
    cout << "Reader thread finished" << endl;
    CloseHandle(hReaderThread);

    /* Wait for the deleter thread */
    WaitForSingleObject(hDeleterThread, INFINITE);
    cout << "Deleter thread finished" << endl;
    CloseHandle(hDeleterThread);

    /* Wait for the inserter thread */
    WaitForSingleObject(hInserterThread, INFINITE);
    cout << "Inserter thread finished" << endl;
    CloseHandle(hInserterThread);

    d_tfsterm(hTFS);
    stop_tfs_process();

    cout << endl << "Press <Enter> to continue . . .";
    getchar();
    return EXIT_SUCCESS;
}
